import { Request, Response } from 'express'
import { container } from '../di/container'

import { OffsetCriteria } from '@/libs/domain/criteria/criteria'
import { Filters } from '@/libs/domain/criteria/filters'
import { CreateCustomerUseCase } from '@/libs/application/use-cases/customer/create.use-case'
import { ShowCustomerUseCase } from '@/libs/application/use-cases/customer/show.use-case'
import { UpdateCustomerUseCase } from '@/libs/application/use-cases/customer/update.use-case'
import { DeleteCustomerUseCase } from '@/libs/application/use-cases/customer/delete.use-case'
import { ListCustomersUseCase } from '@/libs/application/use-cases/customer/list.use-case'
import { OffsetPagination } from '@/libs/domain/criteria/offset-pagination'
import { Order } from '@/libs/domain/criteria/order'
import { RequestContext } from '@/libs/domain/contexts/request.context'

const createCustomerUseCase = container.get(CreateCustomerUseCase)
const showCustomerUseCase = container.get(ShowCustomerUseCase)
const updateCustomerUseCase = container.get(UpdateCustomerUseCase)
const deleteCustomerUseCase = container.get(DeleteCustomerUseCase)
const listCustomersUseCase = container.get(ListCustomersUseCase)

export class CustomerController {
  static async create(req: Request, res: Response) {
    const response = await createCustomerUseCase.run(req.body, RequestContext.fromPrimitives(req))

    return res.status(201).json(response)
  }

  static async update(req: Request, res: Response) {
    const { id } = req.params
    const response = await updateCustomerUseCase.run(
      id,
      req.body,
      RequestContext.fromPrimitives(req)
    )

    return res.json(response)
  }

  static async delete(req: Request, res: Response) {
    const { id } = req.params
    await deleteCustomerUseCase.run(id, RequestContext.fromPrimitives(req))

    return res.sendStatus(204)
  }

  static async show(req: Request, res: Response) {
    const { id } = req.params

    const response = await showCustomerUseCase.run(id, RequestContext.fromPrimitives(req))
    return res.json(response)
  }

  static async list(req: Request, res: Response) {
    const filters = Filters.fromPrimitives(req.query, [
      'user.firstName',
      'user.lastName',
      'user.email',
      'user.phoneNumber',
    ])
    const pagination = OffsetPagination.fromPrimitives(req.query)
    const order = Order.fromPrimitives(req.query, ['user.firstName', 'user.lastName', 'createdAt'])

    const criteria = new OffsetCriteria({
      filters,
      pagination,
      order,
    })
    const response = await listCustomersUseCase.run(criteria, RequestContext.fromPrimitives(req))

    return res.json(response)
  }
}
