import { Request, Response } from 'express'
import { CreateTransportCompanyUseCase } from '@/libs/application/use-cases/transport-company/create.use-case'
import { UpdateTransportCompanyUseCase } from '@/libs/application/use-cases/transport-company/update.use-case'
import { DeleteTransportCompanyUseCase } from '@/libs/application/use-cases/transport-company/delete.use-case'
import { ShowTransportCompanyUseCase } from '@/libs/application/use-cases/transport-company/show.use-case'
import { ListTransportCompaniesUseCase } from '@/libs/application/use-cases/transport-company/list.use-case'
import { Filters } from '@/libs/domain/criteria/filters'
import { OffsetPagination } from '@/libs/domain/criteria/offset-pagination'
import { OffsetCriteria } from '@/libs/domain/criteria/criteria'
import { Order } from '@/libs/domain/criteria/order'
import { RequestContext } from '@/libs/domain/contexts/request.context'
import { container } from '../di/container'

const createTransportCompanyUseCase = container.get(CreateTransportCompanyUseCase)
const updateTransportCompanyUseCase = container.get(UpdateTransportCompanyUseCase)
const deleteTransportCompanyUseCase = container.get(DeleteTransportCompanyUseCase)
const showTransportCompanyUseCase = container.get(ShowTransportCompanyUseCase)
const listTransportCompaniesUseCase = container.get(ListTransportCompaniesUseCase)

export class TransportCompanyController {
  static async create(req: Request, res: Response) {
    const response = await createTransportCompanyUseCase.run(
      req.body,
      RequestContext.fromPrimitives(req)
    )
    return res.status(201).json(response)
  }

  static async update(req: Request, res: Response) {
    const { id } = req.params
    const response = await updateTransportCompanyUseCase.run(
      id,
      req.body,
      RequestContext.fromPrimitives(req)
    )
    return res.status(200).json(response)
  }

  static async delete(req: Request, res: Response) {
    const { id } = req.params
    await deleteTransportCompanyUseCase.run(id, RequestContext.fromPrimitives(req))
    return res.sendStatus(204)
  }

  static async show(req: Request, res: Response) {
    const { id } = req.params
    const transportCompany = await showTransportCompanyUseCase.run(
      id,
      RequestContext.fromPrimitives(req)
    )
    return res.status(200).json(transportCompany)
  }

  static async list(req: Request, res: Response) {
    const filters = Filters.fromPrimitives(req.query, [
      'rut',
      'legalName',
      'legalAddress',
      'email',
      'phoneNumber',
    ])

    const order = Order.fromPrimitives(req.query, [
      'rut',
      'legalName',
      'legalAddress',
      'email',
      'createdAt',
    ])

    const pagination = OffsetPagination.fromPrimitives(req.query)

    const criteria = new OffsetCriteria({
      filters,
      order,
      pagination,
    })
    const response = await listTransportCompaniesUseCase.run(
      criteria,
      RequestContext.fromPrimitives(req)
    )
    return res.status(200).json(response)
  }
}
