import express from 'express'
import morgan from 'morgan'
import cors from 'cors'
import path from 'path'
import type { Server as Httpserver } from 'http'
import { errors as validationErrors } from 'celebrate'

import { logger } from '@/libs/domain/logger'
import { setupErrorTracking } from './api/error-tracking'

import { error400Handler } from './api/error-handlers/400.error-handler'
import { error500Handler } from './api/error-handlers/500.error-handler'
import { environment } from '@/core/api/environment'
import { appDataSource } from '@/libs/infrastructure/persistance/app-data-source'

import { mailerQueue } from '@/libs/infrastructure/queues/mail.queue'

import { apiDocsRouter } from './api-docs/router'

import { apiRouter } from './api/router'

const app = express()
const errorTracking = setupErrorTracking(app)

export class Server {
  private httpServer?: Httpserver

  constructor() {
    this.setViews()
    this.addMiddlewares()
    this.addRouters()
    this.addErrorHandlers()
  }

  async start() {
    await appDataSource.initialize()

    this.httpServer = app.listen(environment.API_PORT, () => {
      logger.debug(`Application is running on port http://localhost:${environment.API_PORT}`)
    })
  }

  setViews() {
    app.set('view engine', 'pug')
    app.set('views', path.join(__dirname, '/../libs/infrastructure/views/pages'))
  }

  addMiddlewares() {
    app.use(...errorTracking.middlewares())
    app.use(morgan(':remote-addr :method :url :status :res[content-length] - :response-time ms'))

    app.use(cors())

    app.use(express.json())
    app.use(express.urlencoded({ extended: true }))
  }

  addRouters() {
    app.use('/api/v1', apiRouter)
    app.use('/docs/v1', apiDocsRouter)
  }

  addErrorHandlers() {
    app.use(errorTracking.errorHandler())
    app.use(validationErrors())
    app.use(error400Handler)
    app.use(error500Handler)
  }

  async close() {
    await Promise.all([mailerQueue.close()])
    await appDataSource.destroy()

    await new Promise((resolve, reject) => {
      this.httpServer?.close(err => {
        if (err) {
          reject(err)
          return
        }
        resolve(true)
      })
    })
  }
}
