import { BaseMasterRepository } from '@/libs/domain/repositories/master/_base-master.repository'
import { EntitySchema, ObjectLiteral, Repository } from 'typeorm'
import {
  FindCriteriaOptions,
  FindOneOptions,
  FindOptions,
  Relations,
  TypeOrmCriteriaParser,
  TypeOrmOptions,
} from '../typeorm-criteria-parser'
import { appDataSource } from '../../app-data-source'
import { DeepPartial } from 'utility-types'
import { CursorCriteria, OffsetCriteria } from '@/libs/domain/criteria/criteria'
import _get from 'lodash/get'
import { BaseMasterEntity } from '@/libs/domain/entities/master/_base-master.entity'

export abstract class TypeOrmBaseMasterRepository<Schema extends BaseMasterEntity & ObjectLiteral>
  implements BaseMasterRepository<Schema>
{
  protected abstract getSchema(): EntitySchema<Schema>

  protected abstract getRelations(): Relations<Schema>

  protected abstract getFindOneOptions(): FindOneOptions<Schema>

  protected abstract getFindOptions(): FindOptions<Schema>

  protected abstract getFindCriteriaOptions(): FindCriteriaOptions<Schema>

  getRepository(): Repository<Schema> {
    return appDataSource.getRepository(this.getSchema())
  }

  async save(entity: DeepPartial<Schema>): Promise<Schema> {
    const _entity = await this.getRepository().save(entity as Schema) // DeepPartial from 'utility-types'  !== DeepPartial from 'typeorm'

    return (await this.findOne({ id: _entity.id! } as Schema))!
  }

  async delete(entity: DeepPartial<Schema>): Promise<void> {
    const _entity = await this.findOne({
      ...entity,
    })
    await this.getRepository().remove(_entity!)
  }

  async findOne(entity: DeepPartial<Schema>): Promise<Schema | null> {
    return this.getRepository().findOne({
      where: entity,
      ...this.getFindOneOptions(),
      relations: this.getRelations(),
    })
  }

  async findOneOrCreate(where: DeepPartial<Schema>, entity: DeepPartial<Schema>): Promise<Schema> {
    let _entity = await this.findOne(where)

    if (_entity) {
      return _entity
    }
    await this.save(entity as Schema)

    _entity = await this.findOne(where)
    return _entity!
  }

  async find(): Promise<Schema[]> {
    return this.getRepository().find({ ...this.getFindOptions(), relations: this.getRelations() })
  }

  async findByCursorCriteria(
    criteria: CursorCriteria
  ): Promise<{ cursor: string | null; results: Schema[] }> {
    let query: TypeOrmOptions = {}

    query = await TypeOrmCriteriaParser.parseCursorCriteria(criteria)

    const results = await this.getRepository().find({
      ...(query as any),
      take: query.take! + 1,
      ...this.getFindCriteriaOptions(),
      relations: this.getRelations(),
    })

    if (results.length === query.take! + 1) {
      const payload = {
        [criteria.pagination.field]: _get(results[query.take! - 1], criteria.pagination.field),
      }

      results.pop()
      return {
        cursor: await criteria.pagination.encodeCursor(payload),
        results,
      }
    }

    return {
      cursor: null,
      results,
    }
  }

  async findByOffsetCriteria(
    criteria: OffsetCriteria
  ): Promise<{ results: Schema[]; total: number }> {
    let query: TypeOrmOptions = {}

    query = await TypeOrmCriteriaParser.parseOffsetCriteria(criteria)
    const [results, total] = await this.getRepository().findAndCount({
      ...(query as any),
      ...this.getFindCriteriaOptions(),
      relations: this.getRelations(),
    })

    return {
      results,
      total,
    }
  }
}
