import { EntitySchema, ObjectLiteral, Repository } from 'typeorm'
import { appDataSource } from '../../app-data-source'
import { DeepPartial } from 'utility-types'
import { CursorCriteria, OffsetCriteria } from '@/libs/domain/criteria/criteria'
import {
  FindCriteriaOptions,
  FindOneOptions,
  FindOptions,
  Relations,
  TypeOrmCriteriaParser,
  TypeOrmOptions,
} from '../typeorm-criteria-parser'

import _get from 'lodash/get'
import { BaseTenantRepository } from '@/libs/domain/repositories/tenant/_base-tenant.repository'
import { BaseTenantEntity } from '@/libs/domain/entities/tenant/_base-tenant.entity'

export abstract class TypeOrmBaseTenantRepository<Schema extends BaseTenantEntity & ObjectLiteral>
  implements BaseTenantRepository<Schema>
{
  protected abstract getSchema(): EntitySchema<Schema>

  protected abstract getRelations(): Relations<Schema>

  protected abstract getFindOneOptions(): FindOneOptions<Schema>

  protected abstract getFindOptions(): FindOptions<Schema>

  protected abstract getFindCriteriaOptions(): FindCriteriaOptions<Schema>

  getRepository(): Repository<Schema> {
    return appDataSource.getRepository(this.getSchema())
  }

  async save(entity: DeepPartial<Schema>, tenantId: string): Promise<Schema> {
    const _entity = await this.getRepository().save({ ...entity, tenantId } as Schema) // DeepPartial from 'utility-types'  !== DeepPartial from 'typeorm'

    return (await this.findOne({ id: _entity.id } as Schema, tenantId))!
  }

  async delete(entity: DeepPartial<Schema>, tenantId: string): Promise<void> {
    const _entity = await this.findOne(
      {
        ...entity,
      },
      tenantId
    )

    await this.getRepository().remove(_entity!)
  }

  async findOne(entity: DeepPartial<Schema>, tenantId: string): Promise<Schema | null> {
    return this.getRepository().findOne({
      where: { ...entity, tenantId },
      ...this.getFindOneOptions(),
      relations: this.getRelations(),
    })
  }

  async findOneOrCreate(
    where: DeepPartial<Schema>,
    entity: DeepPartial<Schema>,
    tenantId: string
  ): Promise<Schema> {
    let _entity = await this.findOne(where, tenantId)

    if (_entity) {
      return _entity
    }
    await this.save(entity as Schema, tenantId)

    _entity = await this.findOne(where, tenantId)
    return _entity!
  }

  async find(tenantId: string): Promise<Schema[]> {
    return this.getRepository().find({
      ...this.getFindOptions(),
      relations: this.getRelations(),
      where: {
        tenantId,
      } as Schema,
    })
  }

  async findByCursorCriteria(
    criteria: CursorCriteria,
    tenantId: string
  ): Promise<{ cursor: string | null; results: Schema[] }> {
    let query: TypeOrmOptions = {}

    query = await TypeOrmCriteriaParser.parseCursorCriteria(criteria)

    query.where = {
      ...query.where,
      tenantId,
    }

    const results = await this.getRepository().find({
      ...(query as any),
      take: query.take! + 1,
      ...this.getFindCriteriaOptions(),
      relations: this.getRelations(),
    })

    if (results.length === query.take! + 1) {
      const payload = {
        [criteria.pagination.field]: _get(results[query.take! - 1], criteria.pagination.field),
      }

      results.pop()
      return {
        cursor: await criteria.pagination.encodeCursor(payload),
        results,
      }
    }

    return {
      cursor: null,
      results,
    }
  }

  async findByOffsetCriteria(
    criteria: OffsetCriteria,
    tenantId: string
  ): Promise<{ results: Schema[]; total: number }> {
    let query: TypeOrmOptions = {}

    query = await TypeOrmCriteriaParser.parseOffsetCriteria(criteria)

    query.where = {
      ...query.where,
      tenantId,
    }

    const [results, total] = await this.getRepository().findAndCount({
      ...(query as any),
      ...this.getFindCriteriaOptions(),
      relations: this.getRelations(),
    })

    return {
      results,
      total,
    }
  }
}
